// Exercise 21.27 Solution
// Non-optimized version.
#include "compiler.h"

int main()
{
   char inFileName[ 15 ] = "", outFileName[ 15 ] = "";
   int last = 0;

   cout << "Enter Simple file to be compiled: ";
   cin >> setw( 15 ) >> inFileName;

   while ( isalnum( inFileName[ last ] ) != 0 ) {
      outFileName[ last ] = inFileName[ last ];
      last++;            // note the last occurance
   }

   outFileName[ last ] = '\0';     // append a NULL character
   strcat( outFileName, ".sml" );  // add .sml to name

   ifstream inFile( inFileName, ios::in );

   if ( inFile )
      compile( inFile, outFileName );
   else
      cerr << "File not opened. Program execution terminating.\n";

   return 0;
}

// compile function calls the first pass and the second pass
void compile( ifstream &input, char *outFileName )
{
   TableEntry symbolTable[ SYMBOLTABLESIZE ]; // symbol table
   int flags[ MEMORYSIZE ];           // array for forward references
   int machineArray[ MEMORYSIZE ];    // array for SML instructions

   initArrays( flags, machineArray, symbolTable );

   firstPass( flags, machineArray, symbolTable, input );
   secondPass( flags, machineArray, symbolTable );

   printOutput( machineArray, outFileName );
}

// firstPass constructs the symbol table, creates SML, and flags unresolved
// references for goto and if/goto statements.
void firstPass( int flags[], int machineArray[], TableEntry symbolTable[],
                ifstream &input )
{
   char array[ MAXIMUM ];              // array to copy a Simple line
   int n = MAXIMUM;                  // required for fgets()
   int dataCounter = MEMORYSIZE - 1; // 1st data location in machineArray
   int instCounter = 0;              // 1st instruction location in machineArray

   input.getline( array, n );

   while ( !input.eof() ) {
      separateToken( array, flags, machineArray, symbolTable, &dataCounter,
                     &instCounter );
      input.getline( array, n );
   }
}

// Separate Tokens tokenizes a Simple statement, process the line number,
// and passes the next token to keyWord for processing.
void separateToken( char array[], int flags[], int machineArray[],
                    TableEntry symbolTable[], int *dataCounterPtr,
                    int *instCounterPtr )
{
   char *tokenPtr = strtok( array, " " );     // tokenize line
   lineNumber( tokenPtr, symbolTable, *instCounterPtr, *dataCounterPtr );
   tokenPtr = strtok( 0, " \n" );      // get next token
   keyWord( tokenPtr, flags, machineArray, symbolTable, dataCounterPtr,
            instCounterPtr );
}

// checkSymbolTable searches the symbol table and returns
// the symbols SML location or a -1 if not found.
int checkSymbolTable( TableEntry symbolTable[], int symbol, char type )
{
   for ( int loop = 0; loop < SYMBOLTABLESIZE; ++loop )
      if ( ( symbol == symbolTable[ loop ].symbol ) && 
           ( type == symbolTable[ loop ].type ) )
         return symbolTable[ loop ].location;  // return SML location

   return -1;                                // symbol not found
}

// lineNumber processes line numbers
void lineNumber( char *tokenPtr, TableEntry symbolTable[],
                 int instCounter, int dataCounter )
{
   const char type = 'L';
   int symbol;

   if ( isdigit( tokenPtr[ 0 ] ) ) {
      symbol = atoi( tokenPtr );

      if ( -1 == checkSymbolTable( symbolTable, symbol, type ) )
         addToSymbolTable( type, symbol, dataCounter, symbolTable, 
                          instCounter );
   }
}

// keyWord determines the key word type and calls the appropriate function
void keyWord( char *tokenPtr, int flags[], int machineArray[],
              TableEntry symbolTable[], int *dataCounterPtr,
              int *instCounterPtr )
{
   if ( strcmp( tokenPtr, "rem" ) == 0 )
      ; // no instructions are generated by comments
   else if ( strcmp( tokenPtr, "input" ) == 0 ) {
      tokenPtr = strtok( 0, " " );  // assign pointer to next token
      keyInput( tokenPtr, machineArray, symbolTable, dataCounterPtr,
                instCounterPtr );
   }
   else if ( strcmp( tokenPtr, "print" ) == 0 ) {
      tokenPtr = strtok( 0, " " );  // assign pointer to next token
      keyPrint( tokenPtr, machineArray, symbolTable, dataCounterPtr,
                instCounterPtr );
   }
   else if ( strcmp( tokenPtr, "goto" ) == 0 ) {
      tokenPtr = strtok( 0, " " );  // assign pointer to next token
      keyGoto( tokenPtr, flags, machineArray, symbolTable, instCounterPtr );
   }
   else if ( strcmp( tokenPtr, "if" ) == 0 ) {
      tokenPtr = strtok( 0, " " );  // assign pointer to next token
      keyIfGoto( tokenPtr, flags, machineArray, symbolTable, dataCounterPtr,
                 instCounterPtr );
   }
   else if ( strcmp( tokenPtr, "end" ) == 0 ) {
      machineArray[ *instCounterPtr ] = HALT * 100;
      ++( *instCounterPtr );
      tokenPtr = 0;       // assign tokenPtr to 0
   }
   else if ( strcmp( tokenPtr, "let" ) == 0 ) {
      tokenPtr = strtok( 0, " " );  // assign pointer to next token
      keyLet( tokenPtr, machineArray, symbolTable, dataCounterPtr,
              instCounterPtr );
   }
}

// keyInput process input keywords
void keyInput( char *tokenPtr, int machineArray[], TableEntry symbolTable[],
               int *dataCounterPtr, int *instCounterPtr )
{
   const char type = 'V';

   machineArray[ *instCounterPtr ] = READ * 100;
   int symbol = tokenPtr[ 0 ];
   int tableTest = checkSymbolTable( symbolTable, symbol, type );

   if ( -1 == tableTest ) {
      addToSymbolTable( type, symbol, *dataCounterPtr, symbolTable,
                        *instCounterPtr );
      machineArray[ *instCounterPtr ] += *dataCounterPtr;
      --( *dataCounterPtr );
   }
   else
      machineArray[ *instCounterPtr ] += tableTest;

   ++( *instCounterPtr );
}

// keyPrint process print keywords
void keyPrint( char *tokenPtr, int machineArray[], TableEntry symbolTable[],
               int *dataCounterPtr, int *instCounterPtr )
{
   const char type = 'V';

   machineArray[ *instCounterPtr ] = WRITE * 100;
   int symbol = tokenPtr[ 0 ];
   int tableTest = checkSymbolTable( symbolTable, symbol, type );

   if ( -1 == tableTest ) {
      addToSymbolTable( type, symbol, *dataCounterPtr, symbolTable,
                        *instCounterPtr );
      machineArray[*instCounterPtr] += *dataCounterPtr;
      --( *dataCounterPtr );
   }
   else
      machineArray[ *instCounterPtr ] += tableTest;

   ++( *instCounterPtr );
}

// keyGoto process goto keywords
void keyGoto( char *tokenPtr, int flags[], int machineArray[],
              TableEntry symbolTable[], int *instCounterPtr )
{
   const char type = 'L';

   machineArray[*instCounterPtr] = BRANCH * 100;
   int symbol = atoi( tokenPtr );
   int tableTest = checkSymbolTable( symbolTable, symbol, type );
   addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
   ++( *instCounterPtr );
}

// keyIfGoto process if/goto commands
void keyIfGoto( char *tokenPtr, int flags[], int machineArray[],
                TableEntry symbolTable[], int *dataCounterPtr, 
                int *instCounterPtr )
{
   int operand1Loc = checkOperand( symbolTable, tokenPtr, dataCounterPtr,
                                   instCounterPtr, machineArray );

   char *operatorPtr = strtok( 0, " " );   // get the operator

   tokenPtr = strtok( 0, " " );  // get the right operand of comparison operator

   int operand2Loc = checkOperand( symbolTable, tokenPtr, dataCounterPtr,
                                   instCounterPtr, machineArray );

   tokenPtr = strtok( 0, " " );  // read in the goto keyword

   char *gotoLinePtr = strtok( 0, " " );  // read in the goto line number

   evaluateExpression( operand1Loc, operand2Loc, operatorPtr, machineArray,
                       symbolTable, instCounterPtr, flags, gotoLinePtr );
}

// checkOperand ensures that the operands of an if/goto statement are
// in the symbol table.
int checkOperand( TableEntry symbolTable[], char *symPtr, int *dataCounterPtr,
                 int *instCounterPtr, int machineArray[] )
{
   char type;
   int tableTest, operand, temp;

   if ( isalpha( symPtr[ 0 ] ) ) {
      type = 'V';
      operand = symPtr[ 0 ];
      tableTest = checkSymbolTable( symbolTable, operand, type );

      if ( tableTest == -1 ) {
         addToSymbolTable( type, operand, *dataCounterPtr, symbolTable,
                           *instCounterPtr );
         temp = *dataCounterPtr;
         --( *dataCounterPtr );
         return temp;
      }
      else
         return tableTest;
   }
   // if the symbol is a digit or a signed digit
   else if ( isdigit( symPtr[ 0 ] ) ||
            ( ( symPtr[ 0 ] == '-' || symPtr[ 0 ] == '+' ) && 
            isdigit( symPtr[ 1 ] ) != 0 ) ) {
      type = 'C';
      operand = atoi( symPtr );
      tableTest = checkSymbolTable( symbolTable, operand, type );

      if ( tableTest == -1 ) {
         addToSymbolTable( type, operand, *dataCounterPtr, symbolTable,
                          *instCounterPtr );
         machineArray[ *dataCounterPtr ] = operand;
         temp = *dataCounterPtr;
         --( *dataCounterPtr );
         return temp ;
      }
      else
         return tableTest;
   }

   return 0;       // default return for compilation purposes
}

// evaluateExpression creates SML for conditional operators
void evaluateExpression( int operator1Loc, int operator2Loc, char *operandPtr,
                         int machineArray[], TableEntry symbolTable[],
                         int *instCounterPtr, int flags[], char *gotoLinePtr )
{
   const char type = 'L';
   int tableTest, symbol;

   if ( strcmp( operandPtr, "==" ) == 0 ) {
      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHZERO * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );
      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );
   }
   else if ( strcmp( operandPtr, "!=" ) == 0 ) {
      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHNEG * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );

      ++( *instCounterPtr );                          

      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHNEG * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );

      ++( *instCounterPtr );
   }
   else if ( strcmp( operandPtr, ">" ) == 0 ) {
      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHNEG * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );
   }
   else if ( strcmp( operandPtr, "<" ) == 0 ) {
      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHNEG * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );
   }
   else if ( strcmp( operandPtr, ">=" ) == 0 ) {
      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHNEG * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHZERO * 100;

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );
   }
   else if ( strcmp( operandPtr, "<=" ) == 0 ) {
      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += operator1Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = SUBTRACT * 100;
      machineArray[ *instCounterPtr ] += operator2Loc;
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHNEG * 100;

      symbol = atoi( gotoLinePtr );
      tableTest = checkSymbolTable( symbolTable, symbol, type );

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );

      machineArray[ *instCounterPtr ] = BRANCHZERO * 100;

      addLineToFlags( tableTest, symbol, flags, machineArray, instCounterPtr );
      ++( *instCounterPtr );
   }
}

// secondPass resolves incomplete SML instructions for forward references
void secondPass( int flags[], int machineArray[], TableEntry symbolTable[] )
{
   const char type = 'L';

   for ( int loop = 0; loop < MEMORYSIZE; ++loop ) {
      if ( flags[ loop ] != -1 ) {
         int symbol = flags[ loop ];
         int flagLocation = checkSymbolTable( symbolTable, symbol, type );
         machineArray[ loop ] += flagLocation;
      }
   }
}

// keyLet processes the keyword let
void keyLet( char *tokenPtr, int machineArray[], TableEntry symbolTable[],
             int *dataCounterPtr, int *instCounterPtr )
{
   const char type = 'V';
   char infixArray[ MAXIMUM ] = "", postfixArray[ MAXIMUM ] = "";
   int tableTest, symbol, location;
   static int subscript = 0;

   symbol = tokenPtr[ 0 ];
   tableTest = checkSymbolTable( symbolTable, symbol, type );

   if ( -1 == tableTest ) {
      addToSymbolTable( type, symbol, *dataCounterPtr, symbolTable,
                        *instCounterPtr );
      location = *dataCounterPtr;
      --( *dataCounterPtr );
   }
   else
      location = tableTest;

   tokenPtr = strtok( 0, " " );     // grab equal sign
   tokenPtr = strtok( 0, " " );     // get next token

   while ( tokenPtr != 0 ) {
      checkOperand( symbolTable, tokenPtr, dataCounterPtr,
                    instCounterPtr, machineArray );
      infixArray[ subscript ] = tokenPtr[ 0 ];
      ++subscript;
      tokenPtr = strtok( 0, " " );  // get next token
   }

   infixArray[ subscript ] = '\0';

   infixToPostfix( infixArray, postfixArray, location, symbolTable,
                   instCounterPtr, dataCounterPtr, machineArray );

   subscript = 0;    // reset static subscript when done
}

void addToSymbolTable( char type, int symbol, int dataCounter,
                       TableEntry symbolTable[], int instCounter )
{
   static int symbolCounter = 0;

   symbolTable[ symbolCounter ].type = type;
   symbolTable[ symbolCounter ].symbol = symbol;

   if ( type == 'V' || type == 'C' )
      symbolTable[ symbolCounter ].location = dataCounter;
   else
      symbolTable[ symbolCounter ].location = instCounter;

   ++symbolCounter;
}

void addLineToFlags( int tableTest, int symbol, int flags[],
                     int machineArray[], const int *instCounterPtr )
{
   if ( tableTest == -1 )
      flags[ *instCounterPtr ] = symbol;
   else
      machineArray[ *instCounterPtr ] += tableTest;
}

void printOutput( const int machineArray[], char *outFileName )
{
   ofstream output( outFileName, ios::out );

   if ( !output )
      cerr << "File was not opened.\n";
   else                   // output every memory cell
      for ( int loop = 0; loop <= MEMORYSIZE - 1; ++loop )
         output << machineArray[ loop ] << '\n';
}

void initArrays( int flags[], int machineArray[], TableEntry symbolTable[] )
{
   TableEntry initEntry = { 0, 0, -1 };

   for ( int loop = 0; loop < MEMORYSIZE; ++loop ) {
      flags[ loop ] = -1;
      machineArray[ loop ] = 0;
      symbolTable[ loop ] = initEntry;
   }
}

//////////////////////////////////////////////////////////////////////////////
// INFIX TO POSTFIX CONVERSION and POSTFIX EVALUATION FOR THE LET STATEMENT //
//////////////////////////////////////////////////////////////////////////////

// infixToPostfix converts an infix expression to a postfix expression
void infixToPostfix( char infix[], char postfix[], int getsVariable,
                     TableEntry symbolTable[], int *instCounterPtr,
                     int *dataCounterPtr, int machineArray[] )
{
   Stack< int > intStack;
   int infixCount, postfixCount, popValue;
   bool higher;
   int leftParen = '(';   /// made int

   // push a left paren onto the stack and add a right paren to infix
   intStack.push( leftParen );
   strcat( infix, ")" );

   // convert the infix expression to postfix
   for ( infixCount = 0, postfixCount = 0; intStack.stackTop(); 
         ++infixCount ) {

      if ( isalnum( infix[ infixCount ] ) )
         postfix[ postfixCount++ ] = infix[ infixCount ];
      else if ( infix[ infixCount ] == '(' )
         intStack.push( leftParen );
      else if ( isOperator( infix[ infixCount ] ) ) {
         higher = true;   // used to store value of precedence test

         while ( higher ) {
            if ( isOperator( static_cast< char >( intStack.stackTop() ) ) )
               if ( precedence( static_cast< char >( intStack.stackTop() ), 
                    infix[ infixCount ] ) )

                  postfix[ postfixCount++ ] = 
                                     static_cast< char >( intStack.pop() );
               else
                  higher = false;
            else
               higher = false;
         }

         // See chapter 21 for a discussion of reinterpret_cast
         intStack.push( reinterpret_cast< int & > ( infix[ infixCount ] ) );
      }
      else if ( infix[ infixCount ] == ')' )
         while ( ( popValue = intStack.pop() ) != '(' )
            postfix[ postfixCount++ ] = static_cast< char >( popValue );
   }

   postfix[ postfixCount ] = '\0';

   evaluatePostfix( postfix, dataCounterPtr, instCounterPtr,
                    machineArray, getsVariable, symbolTable );
}

// check if c is an operator
bool isOperator( char c )
{
   if ( c == '+' || c == '-' || c == '*' || c == '/' || c == '^' )
      return true;
   else
      return false;
}

// If the precedence of operator1 is >= operator2,
bool precedence( char operator1, char operator2 )
{
   if ( operator1 == '^' )
      return true;
   else if ( operator2 == '^' )
      return false;
   else if ( operator1 == '*' || operator1 == '/' )
      return true;
   else if ( operator1 == '+' || operator1 == '-' )
      if ( operator2 == '*' || operator2 == '/' )
         return false;
      else
         return true;

   return false;
}

// evaluate postfix expression and produce code
void evaluatePostfix( char *expr, int *dataCounterPtr,
                      int *instCounterPtr, int machineArray[],
                      int getsVariable, TableEntry symbolTable[] )
{
   Stack< int > intStack;
   int popRightValue, popLeftValue, accumResult, symbolLocation, symbol;
   char type, array[ 2 ] = "";
   int i;

   strcat( expr, ")" );
   
   for ( i = 0; expr[ i ] != ')'; ++i )
      if ( isdigit( expr[ i ] ) ) {
         type = 'C';
         array[ 0 ] = expr[ i ];
         symbol = atoi( array );

         symbolLocation = checkSymbolTable( symbolTable, symbol, type );
         intStack.push( symbolLocation );
      }
      else if ( isalpha( expr[ i ] ) ) {
         type = 'V';
         symbol = expr[ i ];
         symbolLocation = checkSymbolTable( symbolTable, symbol, type );
         intStack.push( symbolLocation );
      }
      else {
         popRightValue = intStack.pop();
         popLeftValue = intStack.pop();
         accumResult = createLetSML( popRightValue, popLeftValue, machineArray,
                                     instCounterPtr, dataCounterPtr, 
                                     expr[ i ] );
         intStack.push( accumResult );
      }

      machineArray[ *instCounterPtr ] = LOAD * 100;
      machineArray[ *instCounterPtr ] += intStack.pop();
      ++( *instCounterPtr );
      machineArray[ *instCounterPtr ] = STORE * 100;
      machineArray[ *instCounterPtr ] += getsVariable;
      ++( *instCounterPtr );
}

int createLetSML( int right, int left, int machineArray[],
                  int *instCounterPtr, int *dataCounterPtr, char oper )
{
   int location;

   switch( oper ) {
      case '+':
         machineArray[ *instCounterPtr ] = LOAD * 100;
         machineArray[ *instCounterPtr ] += left;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = ADD * 100;
         machineArray[ *instCounterPtr ] += right;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = STORE * 100;
         machineArray[ *instCounterPtr ] += *dataCounterPtr;
         location = *dataCounterPtr;
         --( *dataCounterPtr );
         ++( *instCounterPtr );
         return location;
      case '-':
         machineArray[ *instCounterPtr ] = LOAD * 100;
         machineArray[ *instCounterPtr ] += left;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = SUBTRACT * 100;
         machineArray[ *instCounterPtr ] += right;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = STORE * 100;
         machineArray[ *instCounterPtr ] += *dataCounterPtr;
         location = *dataCounterPtr;
         --( *dataCounterPtr );
         ++( *instCounterPtr );
         return location;
      case '/':
         machineArray[ *instCounterPtr ] = LOAD * 100;
         machineArray[ *instCounterPtr ] += left;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = DIVIDE * 100;
         machineArray[ *instCounterPtr ] += right;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = STORE * 100;
         machineArray[ *instCounterPtr ] += *dataCounterPtr;
         location = *dataCounterPtr;
         --( *dataCounterPtr );
         ++( *instCounterPtr );
         return location;
      case '*':
         machineArray[ *instCounterPtr ] = LOAD * 100;
         machineArray[ *instCounterPtr ] += left;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = MULTIPLY * 100;
         machineArray[ *instCounterPtr ] += right;
         ++( *instCounterPtr );
         machineArray[ *instCounterPtr ] = STORE * 100;
         machineArray[ *instCounterPtr ] += *dataCounterPtr;
         location = *dataCounterPtr;
         --( *dataCounterPtr );
         ++( *instCounterPtr );
         return location;
      default:
         cerr << "ERROR: operator not recognized.\n";
         break;
   }

   return 0;    // default return
}
